@@ -466,7 +466,7 @@ static_const struct color_spec color_specs[] = {
 	{"yellowgreen",		0x9ACD32},
 };
 
-#define endof(T) ((T)+sizeof(T)/sizeof(*(T)))
+#define endof(T) ((T) + array_elements(T))
 
 int decode_color(unsigned char *str, struct rgb *col)
 {
@@ -530,15 +530,15 @@ static unsigned char *get_target(unsigned char *a)
 
 void kill_html_stack_item(struct html_element *e)
 {
-	html_format_changed = 1;
-	if (e->dontkill == 2) {
-		internal("trying to kill unkillable element");
-		return;
-	}
 	if (!e || (void *)e == &html_stack) {
 		internal("trying to free bad html element");
 		return;
 	}
+	if (e->dontkill == 2) {
+		internal("trying to kill unkillable element");
+		return;
+	}
+	html_format_changed = 1;
 	if (e->attr.fontface) mem_free(e->attr.fontface);
 	if (e->attr.link) mem_free(e->attr.link);
 	if (e->attr.target) mem_free(e->attr.target);
@@ -1065,6 +1065,16 @@ static void html_font(unsigned char *a)
 	get_color(a, cast_uchar "color", &format_.fg);
 }
 
+static unsigned char *get_url_val_img(unsigned char *a, unsigned char *name)
+{
+	unsigned char *v = get_url_val(a, name);
+	if (v && !v[strcspn(cast_const_char v, "./")]) {
+		mem_free(v);
+		v = NULL;
+	}
+	return v;
+}
+
 static void html_img(unsigned char *a)
 {
 	unsigned char *al;
@@ -1080,9 +1090,8 @@ static void html_img(unsigned char *a)
 		if (format_.link) mem_free(format_.link);
 		if (format_.form) format_.form = NULL;
 		u = join_urls(*al == '#' ? top_href_base() : format_.href_base, al);
-		format_.link = mem_alloc(strlen(cast_const_char u) + 5);
-		strcpy(cast_char format_.link, "MAP@");
-		strcat(cast_char format_.link, cast_const_char u);
+		format_.link = stracpy(cast_uchar "MAP@");
+		add_to_strn(&format_.link, u);
 		format_.attr |= AT_BOLD;
 		mem_free(u);
 		mem_free(al);
@@ -1090,18 +1099,20 @@ static void html_img(unsigned char *a)
 	ismap = format_.link && (F || !has_attr(a, cast_uchar "usemap")) && has_attr(a, cast_uchar "ismap");
 	if (format_.image) mem_free(format_.image), format_.image = NULL;
 	if (
-		(s = get_url_val(a, cast_uchar "data-full")) ||
-		(s = get_url_val(a, cast_uchar "data-normal")) ||
-		(s = get_url_val(a, cast_uchar "data-src")) ||
-		(s = get_url_val(a, cast_uchar "data-defer-src")) ||
-		(s = get_url_val(a, cast_uchar "data-li-src")) ||
-		(s = get_url_val(a, cast_uchar "data-original")) ||
-		(s = get_url_val(a, cast_uchar "data-small")) ||
-		(s = get_url_val(a, cast_uchar "data-lazy")) ||
-		(s = get_url_val(a, cast_uchar "src")) ||
-		(s = get_url_val(a, cast_uchar "dynsrc")) ||
-		(s = get_url_val(a, cast_uchar "data")) ||
-		(s = get_url_val(a, cast_uchar "content"))
+		(s = get_url_val_img(a, cast_uchar "data-full")) ||
+		(s = get_url_val_img(a, cast_uchar "data-src")) ||
+		(s = get_url_val_img(a, cast_uchar "data-normal")) ||
+		(s = get_url_val_img(a, cast_uchar "data-defer-src")) ||
+		(s = get_url_val_img(a, cast_uchar "data-li-src")) ||
+		(s = get_url_val_img(a, cast_uchar "data-original")) ||
+		(s = get_url_val_img(a, cast_uchar "data-small")) ||
+		(s = get_url_val_img(a, cast_uchar "data-lazy")) ||
+		(s = get_url_val_img(a, cast_uchar "data-delay_url")) ||
+		(s = get_url_val_img(a, cast_uchar "src")) ||
+		(s = get_url_val_img(a, cast_uchar "dynsrc")) ||
+		(s = get_url_val_img(a, cast_uchar "data")) ||
+		(s = get_url_val_img(a, cast_uchar "content")) ||
+		(s = get_url_val(a, cast_uchar "src"))
 	    ) {
 		 if (!format_.link && d_opt->braille) goto skip_img;
 		 format_.image = join_urls(format_.href_base, s);
@@ -2070,7 +2081,7 @@ static void new_menu_item(unsigned char *name, long data, int fullname)
 		item->text = name;
 		item->rtext = data == -1 ? cast_uchar ">" : cast_uchar "";
 		item->hotkey = fullname ? cast_uchar "\000\001" : cast_uchar "\000\000"; /* dirty */
-		item->func = data == -1 ? MENU_FUNC do_select_submenu : MENU_FUNC selected_item;
+		item->func = data == -1 ? do_select_submenu : selected_item;
 		item->data = data == -1 ? nmenu : (void *)(my_intptr_t)data;
 		item->in_m = data == -1 ? 1 : 0;
 		item->free_i = 0;
@@ -2098,7 +2109,7 @@ void free_menu(struct menu_item *m) /* Grrr. Recursion */
 	struct menu_item *mm;
 	for (mm = m; mm->text; mm++) {
 		mem_free(mm->text);
-		if (mm->func == MENU_FUNC do_select_submenu) free_menu(mm->data);
+		if (mm->func == do_select_submenu) free_menu(mm->data);
 	}
 	mem_free(m);
 }
@@ -2123,7 +2134,7 @@ static void menu_labels(struct menu_item *m, unsigned char *base, unsigned char
 {
 	unsigned char *bs;
 	for (; m->text; m++) {
-		if (m->func == MENU_FUNC do_select_submenu) {
+		if (m->func == do_select_submenu) {
 			if ((bs = stracpy(base))) {
 				add_to_strn(&bs, m->text);
 				add_to_strn(&bs, cast_uchar " ");
@@ -2139,13 +2150,15 @@ static void menu_labels(struct menu_item *m, unsigned char *base, unsigned char
 
 static int menu_contains(struct menu_item *m, int f)
 {
-	if (m->func != MENU_FUNC do_select_submenu) return (my_intptr_t)m->data == f;
+	if (m->func != do_select_submenu) return (my_intptr_t)m->data == f;
 	for (m = m->data; m->text; m++) if (menu_contains(m, f)) return 1;
 	return 0;
 }
 
-void do_select_submenu(struct terminal *term, struct menu_item *menu, struct session *ses)
+void do_select_submenu(struct terminal *term, void *menu_, void *ses_)
 {
+	struct menu_item *menu = (struct menu_item *)menu_;
+	struct session *ses = (struct session *)ses_;
 	struct menu_item *m;
 	int def = get_current_state(ses);
 	int sel = 0;
@@ -2243,7 +2256,6 @@ static int do_html_select(unsigned char *attr, unsigned char *html, unsigned cha
 		if (preselect == -1 && has_attr(t_attr, cast_uchar "selected")) preselect = order;
 		v = get_exact_attr_val(t_attr, cast_uchar "value");
 		if (!(order & (ALLOC_GR - 1))) {
-			if ((unsigned)order > MAXINT / sizeof(unsigned char *) - ALLOC_GR) overalloc();
 			if ((unsigned)order > MAXINT / sizeof(unsigned char *) - ALLOC_GR) overalloc();
 			val = mem_realloc(val, (order + ALLOC_GR) * sizeof(unsigned char *));
 		}
@@ -2281,6 +2293,7 @@ static int do_html_select(unsigned char *attr, unsigned char *html, unsigned cha
 	*end = en;
 	if (!order) goto abort;
 	fc = mem_calloc(sizeof(struct form_control));
+	if ((unsigned)order > (unsigned)MAXINT / sizeof(unsigned char *)) overalloc();
 	lbls = mem_calloc(order * sizeof(unsigned char *));
 	fc->form_num = last_form_tag ? (int)(last_form_tag - startf) : 0;
 	fc->ctrl_num = last_form_tag ? (int)(attr - last_form_tag) : (int)(attr - startf);
@@ -2672,6 +2685,7 @@ static void html_link(unsigned char *a)
 	}
 	if (
 	    !casecmp(name, cast_uchar "schema", 6) ||
+	    !casecmp(name, cast_uchar "mw-", 3) ||
 	    !casestrcmp(name, cast_uchar "Edit-Time-Data") ||
 	    !casestrcmp(name, cast_uchar "File-List") ||
 	    !casestrcmp(name, cast_uchar "alternate stylesheet") ||
@@ -3321,7 +3335,7 @@ int get_image_map(unsigned char *head, unsigned char *s, unsigned char *eof, uns
 	nm[nmenu].text = label;
 	nm[nmenu].rtext = cast_uchar "";
 	nm[nmenu].hotkey = cast_uchar "";
-	nm[nmenu].func = MENU_FUNC map_selected;
+	nm[nmenu].func = map_selected;
 	nm[nmenu].data = ld;
 	nm[++nmenu].text = NULL;
 	goto se2;
