@@ -66,11 +66,11 @@ METHODDEF(void) my_skip_input_data(j_decompress_ptr cinfo,long num_bytes)
 {
 	if ((unsigned long)num_bytes>cinfo->src->bytes_in_buffer)
 	{
-	 	/* We have to enter skipping state */
-	 	cinfo->src->next_input_byte+=cinfo->src->bytes_in_buffer;
+		/* We have to enter skipping state */
+		cinfo->src->next_input_byte+=cinfo->src->bytes_in_buffer;
 		((struct jpg_decoder *)(global_cimg->decoder))->skip_bytes
-  			=(int)(num_bytes-cinfo->src->bytes_in_buffer);
-  		cinfo->src->bytes_in_buffer=0;
+			=(int)(num_bytes-cinfo->src->bytes_in_buffer);
+		cinfo->src->bytes_in_buffer=0;
 	}
 	else
 	{
@@ -158,10 +158,8 @@ static void gray_to_rgb(unsigned char *data, int pixels)
 /* Fixes returned data in case they are CMYK or grayscale. */
 static inline void fix_data( struct jpg_decoder *deco, int lines_read)
 {
-#ifdef __ICC
-	volatile	/* ICC bug */
-#endif
-	int a;
+	/* ICC bug */
+	icc_volatile int a;
 
 	switch (global_cinfo->output_components){
 		case 1:
@@ -246,10 +244,11 @@ void jpeg_restart(struct cached_image *cimg, unsigned char *data, int length)
 	switch(deco->state){
 		case 0:
 		/* jpeg_read_header */
-	   	if (JPEG_SUSPENDED==jpeg_read_header(global_cinfo,TRUE))
+		if (JPEG_SUSPENDED==jpeg_read_header(global_cinfo,TRUE))
 			break;
 		global_cinfo->buffered_image=TRUE;
 		deco->state=1;
+		/*-fallthrough*/
 
 		case 1:
 		/* If the scaling is sufficiently brutal we can leave out
@@ -327,6 +326,7 @@ void jpeg_restart(struct cached_image *cimg, unsigned char *data, int length)
 		}
 new_scan:
 		deco->state=2;
+		/*-fallthrough*/
 
 		case 2:
 		/* jpeg_start_output */
@@ -352,7 +352,7 @@ susp0:
 					buffer_bytes_per_pixel;
 			}
 
-	 		if ((lines=
+			if ((lines=
 				jpeg_read_scanlines(
 				global_cinfo,deco->scanlines,1))){
 				/* Some lines were written into cimg buffer */
@@ -362,12 +362,13 @@ susp0:
 			}else{
 				/* No lines have been written into cimg
 				 * buffer */
-			 	/* We are suspended and we want more data */
-	 	 		goto susp0; /* Break the outer
+				/* We are suspended and we want more data */
+				goto susp0; /* Break the outer
 					     * switch statement */
 			}
 		}
 		deco->state=4;
+		/*-fallthrough*/
 
 		case 4:
 		/* jpeg_finish_output */
@@ -382,6 +383,7 @@ susp0:
 			goto new_scan;
 		}
 		deco->state=5;
+		/*-fallthrough*/
 
 		case 5:
 		/* jpeg_finish_decompress */
@@ -403,7 +405,13 @@ void jpeg_destroy_decoder(struct cached_image *cimg)
 
 void add_jpeg_version(unsigned char **s, int *l)
 {
+#if defined(LIBJPEG_TURBO_VERSION)
+	add_to_str(s, l, cast_uchar "JPEG-TURBO (");
+	add_to_str(s, l, cast_uchar stringify(LIBJPEG_TURBO_VERSION));
+	add_to_str(s, l, cast_uchar ", ABI ");
+#else
 	add_to_str(s, l, cast_uchar "JPEG (");
+#endif
 #if defined(JPEG_LIB_VERSION_MAJOR) && defined(JPEG_LIB_VERSION_MINOR)
 	add_num_to_str(s, l, JPEG_LIB_VERSION_MAJOR);
 	add_to_str(s, l, cast_uchar ".");
