@@ -294,6 +294,10 @@ int dos_select(int n, fd_set *rs, fd_set *ws, fd_set *es, struct timeval *t, int
 
 #if defined(HAVE_GETADDRINFO) && defined(HAVE_FREEADDRINFO) && !defined(ATHEOS)
 #define USE_GETADDRINFO
+#else
+#ifdef SUPPORT_IPV6
+#undef SUPPORT_IPV6
+#endif
 #endif
 
 #ifdef SUPPORT_IPV6
@@ -345,6 +349,12 @@ int dos_select(int n, fd_set *rs, fd_set *ws, fd_set *es, struct timeval *t, int
 #define sun_volatile
 #endif
 
+#if defined(__TINYC__) && defined(__aarch64__)
+#define tcc_hack	+ 0
+#else
+#define tcc_hack
+#endif
+
 typedef const char *const_char_ptr;
 
 #if defined(BEOS) || (defined(HAVE_BEGINTHREAD) && defined(OS2)) || defined(HAVE_PTHREADS) || (defined(HAVE_ATHEOS_THREADS_H) && defined(HAVE_SPAWN_THREAD) && defined(HAVE_RESUME_THREAD))
