@@ -193,140 +193,11 @@ void cls_redraw_all_terminals(void)
 	foreach(struct terminal, term, lterm, terminals) {
 		if (!F) redraw_terminal_cls(term);
 #ifdef G
-		else {
-			t_resize(term->dev);
-		}
+		else term->dev->resize_handler(term->dev);
 #endif
 	}
 }
 
-#ifdef G
-
-int do_rects_intersect(struct rect *r1, struct rect *r2)
-{
-	return (r1->x1 > r2->x1 ? r1->x1 : r2->x1) < (r1->x2 > r2->x2 ? r2->x2 : r1->x2) && (r1->y1 > r2->y1 ? r1->y1 : r2->y1) < (r1->y2 > r2->y2 ? r2->y2 : r1->y2);
-}
-
-void intersect_rect(struct rect *v, struct rect *r1, struct rect *r2)
-{
-	v->x1 = r1->x1 > r2->x1 ? r1->x1 : r2->x1;
-	v->x2 = r1->x2 > r2->x2 ? r2->x2 : r1->x2;
-	v->y1 = r1->y1 > r2->y1 ? r1->y1 : r2->y1;
-	v->y2 = r1->y2 > r2->y2 ? r2->y2 : r1->y2;
-}
-
-void unite_rect(struct rect *v, struct rect *r1, struct rect *r2)
-{
-	if (!is_rect_valid(r1)) {
-		if (v != r2) memcpy(v, r2, sizeof(struct rect));
-		return;
-	}
-	if (!is_rect_valid(r2)) {
-		if (v != r1) memcpy(v, r1, sizeof(struct rect));
-		return;
-	}
-	v->x1 = r1->x1 < r2->x1 ? r1->x1 : r2->x1;
-	v->x2 = r1->x2 < r2->x2 ? r2->x2 : r1->x2;
-	v->y1 = r1->y1 < r2->y1 ? r1->y1 : r2->y1;
-	v->y2 = r1->y2 < r2->y2 ? r2->y2 : r1->y2;
-}
-
-int is_rect_valid(struct rect *r1)
-{
-	return r1->x1 < r1->x2 && r1->y1 < r1->y2;
-}
-
-#define R_GR	8
-
-struct rect_set *init_rect_set(void)
-{
-	struct rect_set *s;
-	s = mem_calloc(sizeof(struct rect_set) + sizeof(struct rect) * R_GR);
-	s->rl = R_GR;
-	s->m = 0;
-	return s;
-}
-
-void add_to_rect_set(struct rect_set **s, struct rect *r)
-{
-	struct rect_set *ss = *s;
-	int i;
-	if (!is_rect_valid(r)) return;
-	for (i = 0; i < ss->rl; i++) if (!ss->r[i].x1 && !ss->r[i].x2 && !ss->r[i].y1 && !ss->r[i].y2) {
-		x:
-		memcpy(&ss->r[i], r, sizeof(struct rect));
-		if (i >= ss->m) ss->m = i + 1;
-		return;
-	}
-	if ((unsigned)ss->rl > (MAXINT - sizeof(struct rect_set)) / sizeof(struct rect) - R_GR) overalloc();
-	ss = mem_realloc(ss, sizeof(struct rect_set) + sizeof(struct rect) * (ss->rl + R_GR));
-	memset(&(*s = ss)->r[i = (ss->rl += R_GR) - R_GR], 0, sizeof(struct rect) * R_GR);
-	goto x;
-}
-
-void exclude_rect_from_set(struct rect_set **s, struct rect *r)
-{
-	int i, a;
-	struct rect *rr;
-	do {
-		a = 0;
-		for (i = 0; i < (*s)->m; i++) if (do_rects_intersect(rr = &(*s)->r[i], r)) {
-			struct rect r1, r2, r3, r4;
-			r1.x1 = rr->x1;
-			r1.x2 = rr->x2;
-			r1.y1 = rr->y1;
-			r1.y2 = r->y1;
-
-			r2.x1 = rr->x1;
-			r2.x2 = r->x1;
-			r2.y1 = r->y1;
-			r2.y2 = r->y2;
-
-			r3.x1 = r->x2;
-			r3.x2 = rr->x2;
-			r3.y1 = r->y1;
-			r3.y2 = r->y2;
-
-			r4.x1 = rr->x1;
-			r4.x2 = rr->x2;
-			r4.y1 = r->y2;
-			r4.y2 = rr->y2;
-
-			intersect_rect(&r2, &r2, rr);
-			intersect_rect(&r3, &r3, rr);
-			rr->x1 = rr->x2 = rr->y1 = rr->y2 = 0;
-#ifdef DEBUG
-			if (is_rect_valid(&r1) && do_rects_intersect(&r1, r)) internal("bad intersection 1");
-			if (is_rect_valid(&r2) && do_rects_intersect(&r2, r)) internal("bad intersection 2");
-			if (is_rect_valid(&r3) && do_rects_intersect(&r3, r)) internal("bad intersection 3");
-			if (is_rect_valid(&r4) && do_rects_intersect(&r4, r)) internal("bad intersection 4");
-#endif
-			add_to_rect_set(s, &r1);
-			add_to_rect_set(s, &r2);
-			add_to_rect_set(s, &r3);
-			add_to_rect_set(s, &r4);
-			a = 1;
-		}
-	} while (a);
-}
-
-/* memory address r must contain one struct rect
- * x1 is leftmost pixel that is still valid
- * x2 is leftmost pixel that isn't valid any more
- * y1, y2 analogically
- */
-int restrict_clip_area(struct graphics_device *dev, struct rect *r, int x1, int y1, int x2, int y2)
-{
-	struct rect v, rr;
-	rr.x1 = x1, rr.x2 = x2, rr.y1 = y1, rr.y2 = y2;
-	if (r) memcpy(r, &dev->clip, sizeof(struct rect));
-	intersect_rect(&v, &dev->clip, &rr);
-	drv->set_clip_area(dev, &v);
-	return is_rect_valid(&v);
-}
-
-#endif
-
 void draw_to_window(struct window *win, void (*fn)(struct terminal *term, void *), void *data)
 {
 	struct terminal *term = win->term;
@@ -362,7 +233,7 @@ void draw_to_window(struct window *win, void (*fn)(struct terminal *term, void *
 		a = 0;
 		memcpy(&r1, &term->dev->clip, sizeof(struct rect));
 		for (i = 0; i < s->m; i++) if (is_rect_valid(r = &s->r[i])) {
-			drv->set_clip_area(term->dev, r);
+			set_clip_area(term->dev, r);
 			pr(fn(term, data)) {
 #ifdef OOPS
 				return;
@@ -372,10 +243,10 @@ void draw_to_window(struct window *win, void (*fn)(struct terminal *term, void *
 		}
 		if (!a) {
 			struct rect empty = { 0, 0, 0, 0 };
-			drv->set_clip_area(term->dev, &empty);
+			set_clip_area(term->dev, &empty);
 			fn(term, data);
 		}
-		drv->set_clip_area(term->dev, &r1);
+		set_clip_area(term->dev, &r1);
 		mem_free(s);
 #endif
 	}
@@ -397,11 +268,11 @@ ok:
 		struct links_event ev = { EV_REDRAW, 0, 0, 0 };
 		ev.x = term->x;
 		ev.y = term->y;
-		drv->set_clip_area(term->dev, &win->redr);
+		set_clip_area(term->dev, &win->redr);
 		memset(&win->redr, 0, sizeof(struct rect));
 		win->handler(win, &ev, 0);
 	}
-	drv->set_clip_area(term->dev, &term->dev->size);
+	set_clip_area(term->dev, &term->dev->size);
 }
 
 void set_window_pos(struct window *win, int x1, int y1, int x2, int y2)
@@ -640,24 +511,30 @@ struct terminal *init_term(int fdin, int fdout, void (*root_window)(struct windo
 static int process_utf_8(struct terminal *term, struct links_event *ev)
 {
 #if defined(G) || defined(ENABLE_UTF8)
-	if (ev->ev == EV_KBD && ((!F && term_charset(term) == utf8_table)
+	if (ev->ev == EV_KBD) {
+		if ((!F && term_charset(term) == utf8_table)
 #ifdef G
-	    || (F && !(drv->flags & GD_UNICODE_KEYS) && g_kbd_codepage(drv) == utf8_table)
+		    || (F && !(drv->flags & GD_UNICODE_KEYS) && g_kbd_codepage(drv) == utf8_table)
 #endif
-	    )) {
-		size_t l;
-		unsigned char *p;
-		unsigned c;
-		if (ev->x <= 0 || ev->x >= 0x100) goto direct;
-		if ((l = strlen(cast_const_char term->utf8_buffer)) >= sizeof(term->utf8_buffer) - 1 || ev->x < 0x80 || ev->x >= 0xc0)
-			term->utf8_buffer[0] = 0, l = 0;
-		term->utf8_buffer[l] = (unsigned char)ev->x;
-		term->utf8_buffer[l + 1] = 0;
-		p = term->utf8_buffer;
-		GET_UTF_8(p, c);
-		if (!c) return 0;
-		ev->x = c;
-		direct:
+		    ) {
+			size_t l;
+			unsigned char *p;
+			unsigned c;
+			if (ev->x <= 0 || ev->x >= 0x100) goto direct;
+			if ((term->utf8_paste_mode ^ ev->y) & KBD_PASTING) {
+				term->utf8_paste_mode = ev->y & KBD_PASTING;
+				term->utf8_buffer[0] = 0;
+			}
+			if ((l = strlen(cast_const_char term->utf8_buffer)) >= sizeof(term->utf8_buffer) - 1 || ev->x < 0x80 || ev->x >= 0xc0)
+				term->utf8_buffer[0] = 0, l = 0;
+			term->utf8_buffer[l] = (unsigned char)ev->x;
+			term->utf8_buffer[l + 1] = 0;
+			p = term->utf8_buffer;
+			GET_UTF_8(p, c);
+			if (!c) return 0;
+			ev->x = c;
+		}
+direct:
 		term->utf8_buffer[0] = 0;
 	}
 #endif
@@ -742,11 +619,11 @@ void t_resize(struct graphics_device *dev)
 	struct links_event ev = { EV_RESIZE, 0, 0, 0 };
 	term->x = ev.x = dev->size.x2;
 	term->y = ev.y = dev->size.y2;
-	drv->set_clip_area(dev, &dev->size);
+	set_clip_area(dev, &dev->size);
 	foreach(struct window, win, lwin, term->windows) {
 		win->handler(win, &ev, 0);
 	}
-	drv->set_clip_area(dev, &dev->size);
+	set_clip_area(dev, &dev->size);
 }
 
 void t_kbd(struct graphics_device *dev, int key, int flags)
@@ -757,11 +634,14 @@ void t_kbd(struct graphics_device *dev, int key, int flags)
 	r.x2 = dev->size.x2, r.y2 = dev->size.y2;
 	ev.x = key;
 	ev.y = flags;
-	if (upcase(key) == 'L' && flags == KBD_CTRL) {
+	if (upcase(ev.x) == 'L' && !(ev.y & KBD_PASTING) && ev.y & KBD_CTRL) {
 		t_redraw(dev, &r);
 		return;
 	} else {
-		drv->set_clip_area(dev, &r);
+		if (ev.x == KBD_STOP) {
+			abort_background_connections();
+		}
+		set_clip_area(dev, &r);
 		if (list_empty(term->windows)) return;
 		if (ev.x == KBD_CTRL_C || ev.x == KBD_CLOSE) {
 			struct window *prev = list_struct(term->windows.prev, struct window);
@@ -797,7 +677,7 @@ void t_mouse(struct graphics_device *dev, int x, int y, int b)
 	}
 	r.x2 = dev->size.x2, r.y2 = dev->size.y2;
 	ev.x = x, ev.y = y, ev.b = b;
-	drv->set_clip_area(dev, &r);
+	set_clip_area(dev, &r);
 	if (list_empty(term->windows)) return;
 	next = list_struct(term->windows.next, struct window);
 	next->handler(next, &ev, 0);
@@ -871,14 +751,17 @@ static void in_term(void *term_)
 		ev->y -= term->top_margin;
 	}
 	if (ev->ev == EV_KBD || ev->ev == EV_MOUSE) {
-		if (ev->ev == EV_KBD && upcase(ev->x) == 'L' && ev->y & KBD_CTRL) {
+		if (ev->ev == EV_KBD && upcase(ev->x) == 'L' && !(ev->y & KBD_PASTING) && ev->y & KBD_CTRL) {
 			ev->ev = EV_REDRAW;
 			ev->x = term->x;
 			ev->y = term->y;
 			goto send_redraw;
 		}
+		if (ev->ev == EV_KBD && ev->x == KBD_STOP) {
+			abort_background_connections();
+		}
 		if (!list_empty(term->windows)) {
-			if (ev->ev == EV_KBD && ev->x == KBD_CTRL_C) {
+			if (ev->ev == EV_KBD && ev->x == KBD_CTRL_C && !(ev->y & KBD_PASTING)) {
 				struct window *prev = list_struct(term->windows.prev, struct window);
 				prev->handler(prev, ev, 0);
 			} else {
@@ -1027,6 +910,7 @@ static inline char_t utf8_hack(char_t c)
 				SETPOS(cx, y);				\
 			add_to_str(&a, &l, encode_utf_8(c));		\
 			SETPOS(cx + 1, y);				\
+			print_next = 1;					\
 		}							\
 	}								\
 	else if (!c || c == 1) add_chr_to_str(&a, &l, ' ');		\
@@ -1044,6 +928,7 @@ static void redraw_screen(struct terminal *term)
 	int attrib = -1;
 	int mode = -1;
 	int l = 0;
+	int print_next = 0;
 	struct term_spec *s;
 	NO_GFX;
 	if (!term->dirty || (term->master && is_blocked())) return;
@@ -1061,11 +946,16 @@ static void redraw_screen(struct terminal *term)
 				/* make sure that padding is identical */
 				if (chr_has_padding)
 					memcpy(&term->last_screen[p], &term->screen[p], sizeof(chr));
+				if (print_next) {
+					print_next = 0;
+					goto must_print_next;
+				}
 				continue;
 			}
 			/*if ((term->screen[p].at & 0x38) == (term->last_screen[p].at & 0x38) && (term->screen[p].ch == 0 || term->screen[p].ch == 1 || term->screen[p].ch == ' ') && (term->last_screen[p].ch == 0 || term->last_screen[p].ch == 1 || term->last_screen[p].ch == ' ') && (x != term->cx || y != term->cy)) continue;*/
 			/*fprintf(stderr, "%d.%d : %d-%d -> %d-%d\n", x, y, term->last_screen[p].ch, term->last_screen[p].at, term->screen[p].ch, term->screen[p].at);*/
 			memcpy(&term->last_screen[p], &term->screen[p], sizeof(chr));
+			must_print_next:
 #ifdef OPENVMS
 			if (n_chars >= term->x - 6) cy = -1;
 #endif
@@ -1083,6 +973,11 @@ static void redraw_screen(struct terminal *term)
 				goto ppc;
 			}
 		}
+		if (print_next && term->left_margin + term->x < term->real_x) {
+			add_to_str(&a, &l, cast_uchar "\033[0m ");
+			attrib = -1;
+			print_next = 0;
+		}
 	}
 	if (l) {
 		if (s->col) add_to_str(&a, &l, cast_uchar "\033[37;40m");
@@ -1214,26 +1109,27 @@ void set_char(struct terminal *t, int x, int y, unsigned ch, unsigned char at)
 {
 	NO_GFX;
 	t->dirty = 1;
-	if (x >= 0 && x < t->x && y >= 0 && y < t->y) {
+	if ((unsigned)x < (unsigned)t->x && (unsigned)y < (unsigned)t->y) {
 		chr *cc = &t->screen[x + t->x * y];
 		cc->ch = ch;
 		cc->at = at;
 	}
 }
 
-chr *get_char(struct terminal *t, int x, int y)
+const chr *get_char(struct terminal *t, int x, int y)
 {
+	int lx, ly;
 	NO_GFX;
-	if (!t->x || !t->y) {
-		static chr empty;
-		empty.ch = ' ';
-		empty.at = 070;
+	lx = t->x - 1;
+	ly = t->y - 1;
+	if ((lx | ly) < 0) {
+		static_const chr empty = { ' ', 070 };
 		return &empty;
 	}
-	if (x >= t->x) x = t->x - 1;
-	if (x < 0) x = 0;
-	if (y >= t->y) y = t->y - 1;
-	if (y < 0) y = 0;
+	if (x > lx) x = lx;
+	else if (x < 0) x = 0;
+	if (y > ly) y = ly;
+	else if (y < 0) y = 0;
 	return &t->screen[x + t->x * y];
 }
 
@@ -1246,7 +1142,7 @@ void set_color(struct terminal *t, int x, int y, unsigned char c)
 
 void set_only_char(struct terminal *t, int x, int y, unsigned ch, unsigned char at)
 {
-	chr *cc;
+	const chr *cc;
 	NO_GFX;
 	t->dirty = 1;
 	cc = get_char(t, x, y);
@@ -1414,15 +1310,18 @@ void exec_on_terminal(struct terminal *term, unsigned char *path, unsigned char
 		} else {
 			int blockh;
 			unsigned char *param;
+			int paraml;
 			if (is_blocked() && fg) {
 				if (*delet)
 					EINTRLOOP(rs, unlink(cast_const_char delet));
 				return;
 			}
-			param = mem_alloc(strlen(cast_const_char path) + strlen(cast_const_char delet) + 3);
-			param[0] = fg;
-			strcpy(cast_char(param + 1), cast_const_char path);
-			strcpy(cast_char(param + 1 + strlen(cast_const_char path) + 1), cast_const_char delet);
+			param = init_str();
+			paraml = 0;
+			add_chr_to_str(&param, &paraml, fg);
+			add_to_str(&param, &paraml, path);
+			add_chr_to_str(&param, &paraml, 0);
+			add_to_str(&param, &paraml, delet);
 			if (fg == 1) {
 				if (!F) block_itrm(term->fdin);
 #ifdef G
@@ -1435,7 +1334,7 @@ void exec_on_terminal(struct terminal *term, unsigned char *path, unsigned char
 #endif
 			}
 			heap_trim();
-			if ((blockh = start_thread(exec_thread, param, (int)strlen(cast_const_char path) + (int)strlen(cast_const_char delet) + 3, *delet != 0)) == -1) {
+			if ((blockh = start_thread(exec_thread, param, paraml + 1, *delet != 0)) == -1) {
 				if (fg == 1) {
 					if (!F) unblock_itrm(term->fdin);
 #ifdef G
@@ -1465,12 +1364,15 @@ void exec_on_terminal(struct terminal *term, unsigned char *path, unsigned char
 		}
 	} else {
 		unsigned char *data;
-		data = mem_alloc(strlen(cast_const_char path) + strlen(cast_const_char delet) + 4);
-		data[0] = 0;
-		data[1] = fg;
-		strcpy(cast_char(data + 2), cast_const_char path);
-		strcpy(cast_char(data + 3 + strlen(cast_const_char path)), cast_const_char delet);
-		hard_write(term->fdout, data, (int)strlen(cast_const_char path) + (int)strlen(cast_const_char delet) + 4);
+		int datal;
+		data = init_str();
+		datal = 0;
+		add_chr_to_str(&data, &datal, 0);
+		add_chr_to_str(&data, &datal, fg);
+		add_to_str(&data, &datal, path);
+		add_chr_to_str(&data, &datal, 0);
+		add_to_str(&data, &datal, delet);
+		hard_write(term->fdout, data, datal + 1);
 		mem_free(data);
 	}
 }
@@ -1478,10 +1380,12 @@ void exec_on_terminal(struct terminal *term, unsigned char *path, unsigned char
 void do_terminal_function(struct terminal *term, unsigned char code, unsigned char *data)
 {
 	unsigned char *x_data;
+	int x_datal;
 	NO_GFX;
-	x_data = mem_alloc(strlen(cast_const_char data) + 2);
-	x_data[0] = code;
-	strcpy(cast_char(x_data + 1), cast_const_char data);
+	x_data = init_str();
+	x_datal = 0;
+	add_chr_to_str(&x_data, &x_datal, code);
+	add_to_str(&x_data, &x_datal, data);
 	exec_on_terminal(term, NULL, x_data, 0);
 	mem_free(x_data);
 }
