@@ -9,6 +9,8 @@
 #include <stsdef.h>
 #include <descrip.h>
 #include <dvidef.h>
+#include <ttdef.h>
+#include <tt2def.h>
 #include <lib$routines.h>
 #include <libclidef.h>
 
@@ -31,6 +33,14 @@ struct _generic_64 {
 #include <gen64def.h>
 #endif
 
+struct tt_mode {
+	unsigned char class;
+	unsigned char mode;
+	unsigned short width;
+	unsigned basic;
+	unsigned extended;
+};
+
 #include <pthread.h>
 #include <tis.h>
 #ifdef HAVE_SYS_IOCTL_H
@@ -117,71 +127,18 @@ void get_path_to_exe(void)
 	path_to_exe = cast_uchar "links";
 }
 
-static volatile int io_raw = -1;
-
+static int io_raw = -1;
 static int input_handle = -1;
 
-static unsigned old_ctrl_mask = 0;
-
 static $DESCRIPTOR (input_channel_desc, "SYS$INPUT:");
 static unsigned short input_channel;
 
 static pthread_mutex_t io_mutex;
 static pthread_mutex_t cancel_mutex;
 
-static volatile int must_redraw = 0;
-static volatile int ctrl_c = 0;
-static volatile int ctrl_y = 0;
-
 static volatile struct timeval gettimeofday_pool[GETTIMEOFDAY_POOL];
 static volatile unsigned gettimeofday_clock = 0;
 
-static void vms_key_handler(int mode, void (*ast)(void))
-{
-	struct _iosb iosb;
-	int ret;
-	ret = sys$qiow(0, input_channel, IO$_SETMODE | mode, &iosb, 0, 0, ast, 0, 0, 0, 0, 0);
-	if (!$VMS_STATUS_SUCCESS(ret))
-		vms_fatal_exit("sys$qiow failed: %d", ret);
-	if (!$VMS_STATUS_SUCCESS(iosb.iosb$w_status))
-		vms_fatal_exit("IO$M_CTRLCAST failed: %d", iosb.iosb$w_status);
-}
-
-static void ctrl_c_ast(void)
-{
-	if (io_raw <= 0) return;
-	must_redraw = 1;
-	ctrl_c = 1;
-	sys$cancel(input_channel);
-	vms_key_handler(IO$M_CTRLCAST, ctrl_c_ast);
-}
-
-static void ctrl_y_ast(void)
-{
-	if (io_raw <= 0) return;
-	must_redraw = 1;
-	ctrl_y = 1;
-	sys$cancel(input_channel);
-	vms_key_handler(IO$M_CTRLYAST, ctrl_y_ast);
-	vms_key_handler(IO$M_CTRLCAST, ctrl_c_ast);
-}
-
-static void handle_ast_keys(int h)
-{
-	if (ctrl_c) {
-		ctrl_c = 0;
-		hard_write(h, cast_uchar "\003", 1);
-	}
-	if (ctrl_y) {
-		ctrl_y = 0;
-		hard_write(h, cast_uchar "\031", 1);
-	}
-	if (must_redraw) {
-		must_redraw = 0;
-		hard_write(h, cast_uchar "\014", 1);
-	}
-}
-
 #define INPUT_BUFFER_SIZE	16
 
 static void vms_input_thread(void *n, int h)
@@ -210,7 +167,6 @@ static void vms_input_thread(void *n, int h)
 		ret = pthread_mutex_unlock(&cancel_mutex);
 		if (ret)
 			vms_fatal_exit("pthread_mutex_unlock failed: %d", ret);
-		handle_ast_keys(h);
 		ret = pthread_mutex_lock(&io_mutex);
 		if (ret)
 			vms_fatal_exit("pthread_mutex_lock failed: %d", ret);
@@ -230,7 +186,6 @@ static void vms_input_thread(void *n, int h)
 		ret = pthread_mutex_unlock(&io_mutex);
 		if (ret)
 			vms_fatal_exit("pthread_mutex_unlock failed: %d", ret);
-		handle_ast_keys(h);
 		if (vms_ret == SS$_CANCEL ||
 		    vms_ret == SS$_ABORT)
 			continue;
@@ -258,6 +213,28 @@ static void vms_input_thread(void *n, int h)
 	}
 }
 
+static void tty_passthru_mode(int enable)
+{
+	struct _iosb iosb;
+	struct tt_mode mode;
+	int ret;
+
+	ret = sys$qiow(0, input_channel, IO$_SENSEMODE, &iosb, 0, 0, &mode, sizeof(mode), 0, 0, 0, 0);
+	if (!$VMS_STATUS_SUCCESS(ret)) {
+		error("IO$_SENSEMODE failed: %d", ret);
+		return;
+	}
+
+	if (enable) mode.extended |= TT2$M_PASTHRU;
+	else mode.extended &= ~TT2$M_PASTHRU;
+
+	ret = sys$qiow(0, input_channel, IO$_SETMODE, &iosb, 0, 0, &mode, sizeof(mode), 0, 0, 0, 0);
+	if (!$VMS_STATUS_SUCCESS(ret)) {
+		error("IO$_SETMODE failed: %d", ret);
+		return;
+	}
+}
+
 int setraw(int ctl, int save)
 {
 	int ret;
@@ -267,8 +244,9 @@ int setraw(int ctl, int save)
 	if (!$VMS_STATUS_SUCCESS(ret))
 		vms_fatal_exit("sys$assign failed: %d", ret);
 	io_raw = 1;
-	vms_key_handler(IO$M_CTRLCAST, ctrl_c_ast);
-	vms_key_handler(IO$M_CTRLYAST, ctrl_y_ast);
+
+	tty_passthru_mode(1);
+
 	ret = pthread_mutex_unlock(&io_mutex);
 	if (ret)
 		vms_fatal_exit("pthread_mutex_unlock failed: %d", ret);
@@ -296,9 +274,10 @@ void setcooked(int ctl)
 	ret = pthread_mutex_unlock(&cancel_mutex);
 	if (ret)
 		vms_fatal_exit("pthread_mutex_unlock failed: %d", ret);
+
+	tty_passthru_mode(0);
+
 	io_raw = 0;
-	vms_key_handler(IO$M_CTRLCAST, NULL);
-	vms_key_handler(IO$M_CTRLYAST, NULL);
 	ret = sys$dassgn(input_channel);
 	if (!$VMS_STATUS_SUCCESS(ret))
 		vms_fatal_exit("sys$dassgn failed: %d", ret);
@@ -306,7 +285,6 @@ void setcooked(int ctl)
 
 int get_input_handle(void)
 {
-	unsigned int disable_ctrl_mask = LIB$M_CLI_CTRLT | LIB$M_CLI_CTRLY;
 	int ret;
 	if (input_handle >= 0)
 		return input_handle;
@@ -319,7 +297,6 @@ int get_input_handle(void)
 	ret = pthread_mutex_lock(&io_mutex);
 	if (ret)
 		vms_fatal_exit("pthread_mutex_lock failed: %d", ret);
-	lib$disable_ctrl(&disable_ctrl_mask, &old_ctrl_mask);
 	vms_thread_high_priority = 1;
 	input_handle = start_thread(vms_input_thread, NULL, 0, 0);
 	if (input_handle == -1)
@@ -955,10 +932,8 @@ int vms_read(int fd, void *buf, size_t size)
 	int r = vpipe_read(fd, buf, size);
 	if (r != -2) return r;
 #endif
-#ifndef OPENVMS_64BIT
-	return read(fd, buf, size);
-#else
-	{
+#ifdef OPENVMS_64BIT
+	if ((my_uintptr_t)buf + size >= 0x80000000U) {
 		int r;
 		unsigned char static_buffer[BOUNCE_BUFFER_SIZE];
 		unsigned char *bounce_buffer = NULL;
@@ -975,6 +950,7 @@ int vms_read(int fd, void *buf, size_t size)
 		return r;
 	}
 #endif
+	return read(fd, buf, size);
 }
 
 int vms_write(int fd, const void *buf, size_t size)
@@ -983,10 +959,8 @@ int vms_write(int fd, const void *buf, size_t size)
 	int r = vpipe_write(fd, buf, size);
 	if (r != -2) return r;
 #endif
-#ifndef OPENVMS_64BIT
-	return write(fd, buf, size);
-#else
-	{
+#ifdef OPENVMS_64BIT
+	if ((my_uintptr_t)buf + size >= 0x80000000U) {
 		int r;
 		unsigned char static_buffer[BOUNCE_BUFFER_SIZE];
 		unsigned char *bounce_buffer = NULL;
@@ -1003,6 +977,7 @@ int vms_write(int fd, const void *buf, size_t size)
 		return r;
 	}
 #endif
+	return write(fd, buf, size);
 }
 
 void init_os(void)
@@ -1135,8 +1110,6 @@ void os_seed_random(unsigned char **pool, int *pool_size)
 
 void terminate_osdep(void)
 {
-	if (old_ctrl_mask)
-		lib$enable_ctrl(&old_ctrl_mask, NULL);
 }
 
 #else
